unit emiTrackball;
{ TemiTrackball component version 1.00

  Author : Eric MINET - eric.minet@libertysurf.fr

  Status : Freeware with sources

  Target : Delphi 4.0 should work on all 32-bit Delphi and C++ Builder

  See Readme.txt for description

  Revision history :
    1.00 : First public release
}

interface

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, ExtCtrls;

type
  // Current click flag type
  TtbClickAction = (tcaNone, tcaUp, tcaUpLeft, tcaLeft, tcaDownLeft,
    tcaDown, tcaDownRight, tcaRight, tcaUpRight);
  // Increase/Decrease notification event for X and Y values
  TIncrementValueEvent = procedure(Sender : TObject; X, Y : Extended) of object;

  // Trackball component
  TemiTrackball = class(TGraphicControl)
  private
    // Internal fields
    FRepeatTimer     : TTimer;
    FCurrentAction   : TtbClickAction;
    // Property fields
    FIncrementSmall  : Extended;
    FIncrementMedium : Extended;
    FIncrementLarge  : Extended;
    FRepeatDelay     : integer;
    FRepeatPause     : integer;
    FOnIncrement     : TIncrementValueEvent;
    // Property accessors
    procedure SetRepeatDelay(aValue : integer);
    procedure SetRepeatPause(aValue : integer);
  protected
    // Overrided methods
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure Paint; override;
    // Event triggers
    procedure DoIncrement(aX, aY : Extended); virtual;
    // Misc methods
    procedure RepeatTimerExpired(Sender: TObject);
    procedure ProcessClick(X, Y : integer);
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
  published
    // Properties
    property IncrementSmall : Extended read FIncrementSmall write FIncrementSmall;
    property IncrementMedium : Extended read FIncrementMedium write FIncrementMedium;
    property IncrementLarge : Extended read FIncrementLarge write FIncrementLarge;
    property RepeatDelay : integer read FRepeatDelay write SetRepeatDelay;
    property RepeatPause : integer read FRepeatPause write SetRepeatPause;
    // Events
    property OnIncrement : TIncrementValueEvent read FOnIncrement write FOnIncrement;
  end;

implementation

{$R emiTrackballbmp.res}

const
  // The constants depend on the size of the trackball bitmap
  // stored in the resource file. Actual size of the bitmap is 49x49
  // Default control width
  cnTrackballWidth        = 49;
  // Default control height
  cnTrackballHeight       = 49;
  // Minimum radius of sensitivity
  cnTrackballMinRadius    = 3;
  // Maximum radius of sensitivity
  cnTrackballMaxRadius    = 24;
  // Maximum radius of sensitivity for small increment
  cnTrackballSmallRadius  = 10;
  // Maximum radius of sensitivity for medium increment
  cnTrackballMediumRadius = 17;

//////////////////////////////////////////////////////////////////////////////
//                             TemiTrackball                                //
//////////////////////////////////////////////////////////////////////////////
// Object construction
constructor TemiTrackball.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  // General properties initialization
  Width        := cnTrackballWidth;
  Height       := cnTrackballHeight;
  ControlStyle := ControlStyle - [csSetCaption] + [csReplicatable];
  // Internal fields initialization
  FCurrentAction := tcaNone;
  FRepeatTimer   := TTimer.Create(Self);
  FRepeatTimer.Enabled := false;
  FRepeatTimer.OnTimer := RepeatTimerExpired;
  // Properties fields initialization
  FIncrementSmall  := 0.1;
  FIncrementMedium := 1;
  FIncrementLarge  := 10;
  FRepeatDelay     := 400;
  FRepeatPause     := 100;
  FOnIncrement     := nil;
end;

// Object destruction
destructor TemiTrackball.Destroy;
begin
  // Free internal timer
  FRepeatTimer.Free;
  inherited Destroy;
end;

// Write accessor for RepeatDelay property
procedure TemiTrackball.SetRepeatDelay(aValue : integer);
begin
  if (aValue<0) then aValue:=0;
  FRepeatDelay:=aValue;
end;

// Write accessor for RepeatPause property
procedure TemiTrackball.SetRepeatPause(aValue : integer);
begin
  if (aValue<0) then aValue:=0;
  FRepeatPause:=aValue;
end;

// MouseDown
procedure TemiTrackball.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  // If left button is down
  if Button=mbLeft then
  begin
    // Enable the repetition timer with the initial delay before repetition
    FRepeatTimer.Interval := FRepeatDelay;
    FRepeatTimer.Enabled  := True;
    // Process the mouse action
    ProcessClick(X, Y);
  end;
end;

// MouseUp
procedure TemiTrackball.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseUp (Button, Shift, X, Y);
  // Disable repetition timer
  FRepeatTimer.Enabled  := False;
  // If there was an action,
  if FCurrentAction<>tcaNone then
  begin
    // No action
    FCurrentAction:=tcaNone;
    // Refresh the control
    Invalidate;
  end;
end;

// Paint method
procedure TemiTrackball.Paint;
var aBitmap : TBitmap;
    s       : string;
begin
  inherited;
  // Create a temporary Bitmap
  aBitmap := TBitmap.Create;
  try
    // Get the bitmap from the resource using a name
    // specific to the current action
    s:='EMITB_'+IntToStr(integer(FCurrentAction));
    aBitmap.LoadFromResourceName(HInstance, s);
    aBitmap.Transparent:=true;
    aBitmap.TransparentMode:=tmAuto;
    Canvas.Draw(0, 0, aBitmap);
  finally
    aBitmap.Free;
  end;
  // In design mode, draw a dash line rectangle around the control
  if csDesigning in ComponentState then
    with Canvas do
    begin
      Pen.Style := psDash;
      Brush.Style := bsClear;
      Rectangle(0, 0, Width, Height);
    end;
end;

// OnIncrement trigger method
procedure TemiTrackball.DoIncrement(aX, aY : Extended);
begin
  if Assigned(FOnIncrement) then FOnIncrement(Self, aX, aY);
end;

// Repetition timer entry point
procedure TemiTrackball.RepeatTimerExpired(Sender: TObject);
var aPt : TPoint;
begin
  FRepeatTimer.Enabled  := false;
  FRepeatTimer.Interval := FRepeatPause;
  if MouseCapture then
  begin
    try
      GetCursorPos(aPt);
      aPt:=ScreenToClient(aPt);
      ProcessClick(aPt.X, aPt.Y);
      FRepeatTimer.Enabled  := true;
    except
      FRepeatTimer.Enabled := false;
      raise;
    end;
  end;
end;

// Method to process the mouse click
procedure TemiTrackball.ProcessClick(X, Y : integer);
var aRelX, aRelY : integer;
    aDist        : Extended;
    aQuadrant    : Extended;
    aAngle       : Extended;
    aDegre       : integer;
    aAction      : TtbClickAction;
    aIncrement   : Extended;
begin
  aIncrement:=0;
  // Relative coordinates to center of the control
  aRelX:=X-24;
  aRelY:=Y-25;
  // Absolute distance from the center
  aDist:=Sqrt((aRelX*aRelX)+(aRelY*aRelY));
  // If the mouse cursor is within the usable surface
  if (aDist>=cnTrackballMinRadius) and (aDist<=cnTrackballMaxRadius)
  then
    begin
      aRelY:=-aRelY;
      // Calculate the angle between the center and the mouse position
      // to determinate in which quadrant the click occured
      if aRelX >= 0
      then
       if aRelY >= 0
       then aQuadrant := 0
       else aQuadrant := 2 * Pi
      else aQuadrant := Pi;
      if aRelX <> 0
      then aAngle := ArcTan(aRelY / aRelX) + aQuadrant
      else
        if aRelY >= 0
        then aAngle := Pi / 2
        else aAngle := 1.5 * Pi;
      // A pseudo-degree angle (base 800 instead of 360) with a rotation of Pi/8
      aDegre:=Trunc((aAngle+(Pi/8))*400/Pi) mod 800;
      // Find the action equivalent to the angle
      case aDegre of
        0..99    : aAction:=tcaRight;
        100..199 : aAction:=tcaUpRight;
        200..299 : aAction:=tcaUp;
        300..399 : aAction:=tcaUpLeft;
        400..499 : aAction:=tcaLeft;
        500..599 : aAction:=tcaDownLeft;
        600..699 : aAction:=tcaDown;
      else
        aAction:=tcaDownRight;
      end;
      // Increment value is dependant of the radius
      case Trunc(aDist) of
        0..cnTrackballSmallRadius-1 : aIncrement:=FIncrementSmall;
        cnTrackballSmallRadius..cnTrackballMediumRadius-1 : aIncrement:=FIncrementMedium;
      else
        aIncrement:=FIncrementLarge;
      end;
    end
  // Else no current action
  else aAction:=tcaNone;
  // If current action has changed
  if (aAction<>FCurrentAction) then
  begin
    // Set the new action value
    FCurrentAction:=aAction;
    // Refresh the control
    Invalidate;
  end;
  // Trigger the OnIncrement event with appropriate values
  case aAction of
   tcaRight     : DoIncrement(aIncrement, 0);
   tcaUpRight   : DoIncrement(aIncrement, -aIncrement);
   tcaUp        : DoIncrement(0, -aIncrement);
   tcaUpLeft    : DoIncrement(-aIncrement, -aIncrement);
   tcaLeft      : DoIncrement(-aIncrement, 0);
   tcaDownLeft  : DoIncrement(-aIncrement, aIncrement);
   tcaDown      : DoIncrement(0, aIncrement);
   tcaDownRight : DoIncrement(aIncrement, aIncrement);
  end;
end;

end.
